<!DOCTYPE html>
<html lang="{{ site.lang | default: "en-US" }}">
  <head>
    <meta charset='utf-8'>
    <meta http-equiv="X-UA-Compatible" content="chrome=1">
    <link href='https://fonts.googleapis.com/css?family=Chivo:900' rel='stylesheet' type='text/css'>
    <link rel="stylesheet" href="{{ '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}">
    <link rel="stylesheet" type="text/css" href="{{ '/assets/css/print.css' | relative_url }}" media="print">
    <!--[if lt IE 9]>
    <script src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script>
    <![endif]-->
    <title>{{ site.title | default: site.github.repository_name }} by {{ site.github.owner_name }}</title>
  </head>

  <body>
    <div id="container">
      <div class="inner">

        <header>
          <p align="center">
            <img src="https://yunjunzhang.files.wordpress.com/2017/08/logo_githubpage.pdf">
          </p>
        </header>
        <section id="downloads" class="clearfix">
          {% if site.github.is_project_page %}
          <a href="{{ site.github.repository_url }}" class="button"><span style="text-align:center">View on GitHub</span></a>
          {% else %}
          <a href="{{ site.github.owner_url }}" class="button"><span style="text-align:center">View on GitHub</span></a>
          {% endif %}
          <a href="{{ site.github.wiki_url }}" class="button"><span style="text-align:center">Wiki</span></a>
          {% if site.show_downloads %}
          <a href="{{ site.github.releases_url }}" class="button"><span style="text-align:center">Download</span></a>
          {% endif %}
        </section>
        <hr>
        <section id="main_content">
          {{ content }}
        </section>

        <footer>
        {% if site.github.is_project_page %}
          {{ site.title | default: site.github.repository_name }} is maintained by <a href="{{ site.github.owner_url }}">{{ site.github.owner_name }}</a><br>
        {% endif %}
          This page was generated by <a href="https://pages.github.com">GitHub Pages</a>.
        </footer>

      </div>
    </div>

    {% if site.google_analytics %}
      <script type="text/javascript">
        (function(i,s,o,g,r,a,m){i['GoogleAnalyticsObject']=r;i[r]=i[r]||function(){
        (i[r].q=i[r].q||[]).push(arguments)},i[r].l=1*new Date();a=s.createElement(o),
        m=s.getElementsByTagName(o)[0];a.async=1;a.src=g;m.parentNode.insertBefore(a,m)
        })(window,document,'script','//www.google-analytics.com/analytics.js','ga');
        ga('create', '{{ site.google_analytics }}', 'auto');
        ga('send', 'pageview');
      </script>
    {% endif %}
  </body>
</html>
