<!DOCTYPE html>
<html lang="{{ site.lang | default: "en-US" }}">
  {% include head.html %}
  <body>
    {% include page-header.html %}
    <section class="main-content">
      {{ content }}
      {% include page-footer.html %}
    </section>
  </body>
</html>
