<head>
  <meta charset="UTF-8">
  <title>{% if site.github.owner_name %}{{ site.github.owner_name }} | {% endif %}{{ site.title | default: site.github.repository_name }}</title>
  <meta name="description" content="{{ page.description | default: site.description | default: site.github.project_tagline }}"/>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <meta name="theme-color" content="#157878">
  <link href='https://fonts.googleapis.com/css?family=Open+Sans:400,700' rel='stylesheet' type='text/css'>
  <link rel="stylesheet" href="{{ '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}">
  {% if site.google_analytics and jekyll.environment == 'production' %}
  {% include analytics.html %}
  {% endif %}
 </head>
