<section class="page-header">
  <h1 class="project-name">
    {{ site.title | default: site.github.repository_name }}
  </h1>
  <h2 class="project-tagline">{{ site.description | default: site.github.project_tagline }}</h2>
  {% if site.show_downloads %}
    <a href="{{ site.github.releases_url }}" class="btn">Download</a>
  {% endif %}
  <a href="https://mintpy.readthedocs.io/" class="btn">Documentation</a>
  {% if site.github.is_project_page %}
    <a href="{{ site.github.repository_url }}" class="btn">View on GitHub</a>
  {% else %}
    <a href="{{ site.github.owner_url }}" class="btn">View on GitHub</a>
  {% endif %}
</section>
