##Figure Setting
font_name = 'Helvetica'
font_size = 12
marker_size = 5
edge_width = 1.0
transparency = 0.8

#AGU recommend .eps for vector, .tiff for raster and .jpg for large raster
#Google Doc support .jpg, .png and .gif only
fig_ext = '.jpg'
fig_dpi = 600
